//Wir wollen in der lage sein commandozeilen parameter zu verarbeiten
#include <clang/Tooling/CommonOptionsParser.h>
//Wir sind eine frontend action
#include <clang/Frontend/FrontendActions.h>
//Wir machen Tooling
#include "clang/Tooling/Tooling.h"
//Wir benutzen die Visitor strategy
#include <clang/AST/RecursiveASTVisitor.h>
//Wir müssen mit dem compiler interagieren können.
#include <clang/Frontend/CompilerInstance.h>
#include <spdlog/spdlog.h>
#include <iostream>
#include "CTUApexAction.h"
#include "CTUApexConsumer.h"

#include "MCGReader.h"
#include "MCGWriter.h"

#include "Util.h"
#include "CallSites.h"
#include "ExtractionMetadata.h"

class FilePropertiesMetaData : public metacg::MetaData::Registrar<FilePropertiesMetaData> {
public:
    static constexpr const char *key = "fileProperties";
    FilePropertiesMetaData() : origin("INVALID"), fromSystemInclude(false), lineNumber(0) {}
    explicit FilePropertiesMetaData(const nlohmann::json &j) {
        if (j.is_null()) {
            metacg::MCGLogger::instance().getConsole()->trace("Could not retrieve meta data for {}", key);
            return;
        }

        std::string fileOrigin = j["origin"].get<std::string>();
        bool isFromSystemInclude = j["systemInclude"].get<bool>();
        origin = fileOrigin;
        fromSystemInclude = isFromSystemInclude;
    }

    nlohmann::json to_json() const final {
        nlohmann::json j;
        j["origin"] = origin;
        j["systemInclude"] = fromSystemInclude;
        return j;
    };

    virtual const char *getKey() const final { return key; }

    std::string origin;
    bool fromSystemInclude;
    int lineNumber;
};
class PiraOneData : public metacg::MetaData::Registrar<PiraOneData> {
public:
    static constexpr const char *key = "numStatements";
    PiraOneData() = default;
    explicit PiraOneData(const nlohmann::json &j) {
        metacg::MCGLogger::instance().getConsole()->trace("Running PiraOneMetaDataRetriever::read from json");
        if (j.is_null()) {
            metacg::MCGLogger::instance().getConsole()->trace("Could not retrieve meta data for {}", key);
            return;
        }
        auto jsonNumStmts = j.get<long long int>();
        metacg::MCGLogger::instance().getConsole()->debug("Read {} stmts from file", jsonNumStmts);
        setNumberOfStatements(jsonNumStmts);
    }

    nlohmann::json to_json() const final {
        nlohmann::json j;
        j["numStatements"] = getNumberOfStatements();
        return j;
    };

    virtual const char *getKey() const final { return key; }

    void setNumberOfStatements(int numStmts) { this->numStmts = numStmts; }
    int getNumberOfStatements() const { return this->numStmts; }
    void setHasBody(bool hasBody = true) { this->hasBody = hasBody; }
    bool getHasBody() const { return this->hasBody; }
    void setDominantRuntime(bool dominantRuntime = true) { this->dominantRuntime = dominantRuntime; }
    bool isDominantRuntime() const { return this->dominantRuntime; }
    void setComesFromCube(bool fromCube = true) { this->wasInPreviousProfile = fromCube; }
    bool comesFromCube() const { return this->wasInPreviousProfile; }
    bool inPreviousProfile() const { return wasInPreviousProfile; }

private:
    bool wasInPreviousProfile{false};
    bool dominantRuntime{false};
    bool hasBody{false};
    int numStmts{0};
};

inline void to_json(nlohmann::json &j, const PiraOneData &data) {
    j = nlohmann::json{{"numStatements", data.getNumberOfStatements()}};
}

//Das sind convenience definitionen, dass wir nicht so viel tippen müssen
using namespace llvm;
using namespace clang;
using namespace clang::tooling;
using namespace clang;

//Die commandozeilen optionen die wir parsen wollen, gehören zu unserem program
static cl::OptionCategory CTUApex("Cross Translation Unit App Extraction");
//static cl::opt<std::string> callsites_file_opt("callsites-file", cl::desc("Config file with the callsites to outline"),
//                                               cl::value_desc("filename"), cl::Required, cl::cat(CTUApex));
static cl::opt<std::string> cg_file("cg-file", cl::desc("MetaCG file containing a whole programm call graph"),
                                    cl::value_desc("filename"), cl::Required, cl::cat(CTUApex));
static cl::opt<std::string> output_file("output", cl::desc("Specify output filename"), cl::value_desc("filename"),
                                        cl::Required, cl::cat(CTUApex));
enum class LogLevel {
    Trace, Debug, Info, Warning, Error, /* Critical,*/ Off
};
static cl::opt<LogLevel> LoggingLevel("log-level", cl::desc("Select log level"),
                                      cl::values(clEnumValN(LogLevel::Trace, "trace", "Enable Trace and higher logs"),
                                                 clEnumValN(LogLevel::Debug, "debug", "Enable Debug and higher logs"),
                                                 clEnumValN(LogLevel::Info, "info", "Enable Info and higher logs"),
                                                 clEnumValN(LogLevel::Warning, "warning",
                                                            "Enable Warning and higher logs"),
                                                 clEnumValN(LogLevel::Error, "error", "Enable only Error logs"),
                                                 clEnumValN(LogLevel::Off, "off", "Disable all logging")),
                                      cl::init(LogLevel::Info), cl::cat(CTUApex));
static cl::opt<bool> OnlyWrapper("generate-only-wrapper", cl::desc("Output only the wrapper function"),
                                 cl::cat(CTUApex));
static cl::opt<bool> DumpAnalysisResult("dump-analysis-result",
                                        cl::desc("Dump the dependence analysis result to stdout"), cl::cat(CTUApex));



template<typename T, class... Args>
std::unique_ptr<clang::tooling::FrontendActionFactory> argumentParsingFrontendActionFactory(Args... args) {

    class SimpleFrontendActionFactory : public clang::tooling::FrontendActionFactory {
    public:
        explicit SimpleFrontendActionFactory(Args... constructorArgs) {
            fn1 = [=]() { return std::make_unique<T>(constructorArgs...); };
        }

        std::unique_ptr<clang::FrontendAction> create() override {
            return fn1();
        }

    private:
        std::function<std::unique_ptr<T>()> fn1; // function
    };

    return std::unique_ptr<FrontendActionFactory>(new SimpleFrontendActionFactory(args...));
}

//Der einsprungspunt in das programm
int main(int argc, const char **argv) {
    //Wir wollen commandozeilen parameter verarbeiten können, also nehmen wir einen option parser
    //Da beim parsen etwas schief gehen kann, machen wir den extra schritt mit expectation

    auto ExpectedParser = CommonOptionsParser::create(argc, argv, CTUApex);

    if (!ExpectedParser) {
        // Fail gracefully for unsupported options.
        llvm::errs() << ExpectedParser.takeError();
        return 1;
    }


    switch (LoggingLevel) {
        case LogLevel::Trace:
            spdlog::set_level(spdlog::level::trace);
            break;
        case LogLevel::Debug:
            spdlog::set_level(spdlog::level::debug);
            break;
        case LogLevel::Info:
            spdlog::set_level(spdlog::level::info);
            break;
        case LogLevel::Warning:
            spdlog::set_level(spdlog::level::warn);
            break;
        case LogLevel::Error:
            spdlog::set_level(spdlog::level::err);
            break;
        case LogLevel::Off:
            spdlog::set_level(spdlog::level::off);
            break;
    }

    //Das ist der eigentliche option parser
    CommonOptionsParser &OptionsParser = ExpectedParser.get();
    //Wir erstellen ein clang tool objekt
    ClangTool Tool(OptionsParser.getCompilations(), OptionsParser.getSourcePathList());
    std::string error="error";
    auto fcd = FixedCompilationDatabase::loadFromDirectory("",error);
    metacg::io::FileSource fs(cg_file.getValue());
    metacg::io::VersionTwoMetaCGReader mcgReader(fs);
    auto& mcgManager=metacg::graph::MCGManager::get();
    mcgManager.addToManagedGraphs("emptyGraph",std::make_unique<metacg::Callgraph>());
    mcgReader.read(mcgManager);

    //Wir führen die frontend action mit unserem tool aus
    Tool.run(argumentParsingFrontendActionFactory<CTUApexAction>(mcgManager.getCallgraph()).get());
    return 0;
}














