
#include <llvm/Support/CommandLine.h>

#include <clang/Basic/Diagnostic.h>

#include <clang/Tooling/CommonOptionsParser.h>
#include <clang/Tooling/Tooling.h>

#include <clang/Frontend/CompilerInstance.h>

#include <clang/Lex/Preprocessor.h>
#include <clang/Lex/PPCallbacks.h>

#include <algorithm>
#include <iostream>
#include <string>
#include <vector>
#include <utility>

#include <clang/Frontend/FrontendActions.h>

namespace path_utils {

    std::string extractTail(std::string path) {
        // normalize slashes
        std::replace(path.begin(), path.end(), '\\', '/');
        // leave only trailing path component
        std::string::size_type tailPos = path.rfind('/');
        tailPos = (tailPos == std::string::npos) ? 0UL : (tailPos + 1);
        return std::string(path, tailPos);
    }

    std::pair<std::string, std::string> crackName(const std::string &name) {
        const std::string::size_type dotPos = name.rfind('.');
        const std::string::size_type extPos = (dotPos == std::string::npos) ? 0UL : (dotPos + 1);
        return std::make_pair(std::string(name, 0UL, dotPos), std::string(name, extPos));
    }

    std::string extractRoot(const std::string &name) {
        return std::string(name, 0UL, name.rfind('.'));
    }

}


static std::vector<std::string> getKnownHdrExts() {
    return {"h", "H", "hpp", "HPP", "hxx", "HXX"};
}

template<typename C, typename T>
inline bool contains(const C &c, const T &val) {
    return std::find(c.begin(), c.end(), val) != c.end();
}

class IncludeFinder : public clang::PPCallbacks {
public:
    explicit inline IncludeFinder(const clang::CompilerInstance &compiler)
            : compiler(compiler) {
        const clang::FileID mainFile = compiler.getSourceManager().getMainFileID();
        name = compiler.getSourceManager().getFileEntryForID(mainFile)->getName();
    }

public:
    inline std::unique_ptr<clang::PPCallbacks> createPreprocessorCallbacks();

    inline void diagnoseAndReport() {
        IncludeInfo selfInclude;

        const std::string &tail = path_utils::extractTail(name);
        const std::string &root = path_utils::extractRoot(tail);

        typedef Includes::iterator It;
        for (auto &include: includes) {
            const std::string &hdrPath = include.second;
            const std::string &hdrName = path_utils::extractTail(hdrPath);

            const std::pair<std::string, std::string> nameParts =
                    path_utils::crackName(hdrName);

            // if root parts are the same and extension is one of known
            // TODO: maybe compare roots case insensitive
            if (nameParts.first == root) {
                if (contains(getKnownHdrExts(), nameParts.second)) {
                    if (selfInclude.second.empty()) {
                        selfInclude = include;
                    } else {
                        // TODO: issue a warning about failed self include detection
                        std::cout << name << ':' << "ambiguous header name detection: " << hdrPath << std::endl;
                    }
                } else {
                    std::cout << name << ':' << "unexpected extension: " << nameParts.second << std::endl;
                }
            }
        }

        if (!selfInclude.second.empty() && !includes.empty()) {
            if (includes[0] != selfInclude) {
                std::cout << name << ':' << selfInclude.first << ':' << "should be the first include in the file"
                          << std::endl;
            }
        }
    }

    inline void InclusionDirective(clang::SourceLocation HashLoc, const clang::Token &IncludeTok,
                                   clang::StringRef FileName, bool IsAngled,
                                   clang::CharSourceRange FilenameRange,
                                   clang::Optional<clang::FileEntryRef> File,
                                   clang::StringRef SearchPath, clang::StringRef RelativePath,
                                   const clang::Module *Imported,
                                   clang::SrcMgr::CharacteristicKind FileType) override {
        clang::SourceManager &sm = compiler.getSourceManager();
        std::cout << "Found inclusion directive\n";
        std::cout << "Hashloc\n";
        HashLoc.dump(sm);
        std::cout << "Could not dump include Tok\n";
        std::cout << "FileName:\n";
        std::cout << FileName.str() << "\n";
        std::cout << "Is Angled:\n";
        std::cout << IsAngled << "\n";
        std::cout << "CharSourceRange\n";
        FilenameRange.getBegin().dump(sm);
        FilenameRange.getEnd().dump(sm);
        std::cout << "Optional File:\n";
        std::cout << File.hasValue() << "\n";
        if (File.hasValue()) {
            std::cout << File.value().getName().str();
        }
        std::cout << "SearchPath:\n";
        std::cout << SearchPath.str() << "\n";
        std::cout << "RelativePath:\n";
        std::cout << RelativePath.str() << "\n";
        std::cout << "Module:\n";
        if (Imported != nullptr) {
            Imported->dump();
        } else {
            std::cout << "Module Imported pointer was nullptr\n";
        }
        std::cout << "FileType:\n";
        std::cout << FileType << "\n";


        if (sm.isInMainFile(HashLoc)) {
            const unsigned int lineNum = sm.getSpellingLineNumber(HashLoc);
            includes.emplace_back(lineNum, FileName.str());
        }
    }

private:
    const clang::CompilerInstance &compiler;
    std::string name;

    typedef std::pair<int, std::string> IncludeInfo;
    typedef std::vector<IncludeInfo> Includes;
    Includes includes;
};

class IncludeFinderAction : public clang::PreprocessOnlyAction {
protected:
    void ExecuteAction() override {

        std::cout << __LINE__ << "\n";
        IncludeFinder includeFinder(getCompilerInstance());
        getCompilerInstance().getPreprocessor().addPPCallbacks(includeFinder.createPreprocessorCallbacks());
        clang::PreprocessOnlyAction::ExecuteAction();
        includeFinder.diagnoseAndReport();
    }
};

class CallbacksProxy : public clang::PPCallbacks {
public:
    inline explicit CallbacksProxy(clang::PPCallbacks &master) : master(master) {}

public:
    void InclusionDirective(clang::SourceLocation HashLoc, const clang::Token &IncludeTok,
                            clang::StringRef FileName, bool IsAngled,
                            clang::CharSourceRange FilenameRange, clang::Optional<clang::FileEntryRef> File,
                            clang::StringRef SearchPath, clang::StringRef RelativePath,
                            const clang::Module *Imported,
                            clang::SrcMgr::CharacteristicKind FileType) override {
        std::cout << __LINE__ << "\n";
        master.InclusionDirective(HashLoc, IncludeTok, FileName, IsAngled, FilenameRange, File, SearchPath,
                                  RelativePath, Imported, FileType);
    }

private:
    clang::PPCallbacks &master;
};

inline std::unique_ptr<clang::PPCallbacks> IncludeFinder::createPreprocessorCallbacks() {
    return std::make_unique<CallbacksProxy>(*this);
}

using namespace clang::tooling;

static llvm::cl::OptionCategory toolCategory("self-inc-first options");

static llvm::cl::extrahelp commonHelp(CommonOptionsParser::HelpMessage);

class : public clang::DiagnosticConsumer {
public:
    bool IncludeInDiagnosticCounts() const override {
        return false;
    }
} diagConsumer;

int main(int argc, const char *argv[]) {

    auto optionsParser = CommonOptionsParser::create(argc, argv, toolCategory);
    ClangTool tool(optionsParser->getCompilations(), optionsParser->getSourcePathList());


    tool.setDiagnosticConsumer(&diagConsumer);

    return tool.run(newFrontendActionFactory<IncludeFinderAction>().get());
}